# coding: utf-8
"""
Celery таски приложения Фед. отчетность
"""
__author__ = 'fattakhov'
import datetime
import logging
import traceback

from celery.result import states
import settings as gs_settings

from m3.actions import ApplicationLogicException

from kinder.core.async.tasks import AsyncTask
from kinder.core.async.models import RunningTask
from kinder.core.unit.models import Unit, UnitKind
from kinder.core.helpers import get_current_user

from exceptions import ReportGenerationException, SOAPClientCreateException
from helpers import ReportFormHelper, SendReportException
from models import ReportForm, ReportFormRow, ScheduleSettings, GisdoUnit
from reports import ReportFormReport

log = logging.getLogger('gisdo')


def _get_report_progress(current_unit_num, units_count):
        """
        Подсчет прогресса выполнения
        @current_unit_num - порядковый номер текущего учреждения
        @units_count - количество учреждений
        """

        return int(100*float(current_unit_num)/float(units_count))


class Report(AsyncTask):
    """
    Отчет по собранным показателям
    """

    description = u'Отчет по собранным показателям'
    stop_executing = False
    state = {'values': {}, 'description': u'Неизвестно', 'progress': u'Неизвестно'}

    def run(self, result_name, report_id, full_name, email, phone):
        """Выполнение задачи"""

        self.set_progress(u'Формирование отчета...', task_state=u'RUNNING')

        self.result_name = result_name
        report = ReportFormReport(result_name)

        report.flush_header(full_name, email, phone)

        try:
            report_form = ReportForm.objects.get(id=report_id)
        except ReportForm.DoesNotExist:
            raise ReportGenerationException(u'Отсутствуют данные по учреждению')

        report_form_rows = ReportFormRow.objects.filter(report=report_form)
        rows_count = report_form_rows.count()

        row_number = 0
        row_data = []
        for report_form_row in report_form_rows:
            try:
                not_on_federal_report = report_form_row.unit.gisdo.not_on_federal_report
            except GisdoUnit.DoesNotExist:
                raise ApplicationLogicException(
                    u'Учреждение c ID=%d не найдено в модуле "Фед. отчетности"' % report_form_row.unit.pk)
            if report_form_row.unit.kind.id == UnitKind.DOU and not not_on_federal_report:

                self.set_progress(u'Пeчать показателей по %s. (%s%%)' % (report_form_row.unit.display(),
                                                                          _get_report_progress(row_number, rows_count)))

                parent_unit = report_form_row.unit.get_mo() or report_form.unit
                ocato = parent_unit.ocato if parent_unit else ''
                try:
                    row_data.append(report.collect_row_data(report_form_row, report_form.date, ocato))
                except (KeyError, UnicodeDecodeError):
                    err_msg = u'При печате показателей по %s произошла ошибка' % report_form_row.unit.display()
                    raise KeyError(err_msg)

                row_number += 1

        report.flush_row_data(sorted(row_data, key=lambda x: x['mo']))
        report.build()

        return self.state

    def on_success(self, retval, task_id, args, kwargs):

        url = '%s/%s.xlsx' % (gs_settings.DOWNLOADS_URL, self.result_name)
        self.set_progress(u'<a href="%s" target="_blank">Отчет сформирован</a>' % url, task_state=states.SUCCESS)

    def on_failure(self, exc, task_id, args, kwargs, einfo):

        self.set_progress(exc.message, task_state=states.FAILURE)


class CollectUnitsDataTask(AsyncTask):
    """
    Сбор и сохранение данных по учреждениям
    """
    description = u'Подсчет показателей'
    stop_executing = False
    state = {'values': {}, 'description': u'Неизвестно', 'progress': u'Неизвестно'}

    EMPTY_MO_MESSAGE = u'Учреждение %s не было добавлено в отчет так как оно не является ДОУ и не имеет дочерних учреждений'
    SUCCESS_COLLECT_MO_DATA = u'Данные по %s успешно собраны'
    FAILURE_COLLECT_MO_DATA = u'Во время сбора данных по %s произошла ошибка'
    PREPARE_DATA_MESSAGE = u'Подготовка данных завершена(%s)'
    NOT_FOUND_MESSAGE = u'Учреждение с id=%i не найдено'

    def _ui_log(self, mo_name, has_error, msg, success_message=SUCCESS_COLLECT_MO_DATA,
                failure_message=FAILURE_COLLECT_MO_DATA):

        msg_status = success_message % mo_name if not has_error else failure_message % mo_name
        msg_detail = '' if not has_error else msg

        self.set_progress(values={msg_status: msg_detail})

    def run(self, context, **kwargs):
        """Выполнение задачи"""
        self.set_progress(u'Подсчет показателей...', task_state=u'RUNNING')

        report_helper = ReportFormHelper(context.report_form, ScheduleSettings.get_settings())
        self.set_progress(values={CollectUnitsDataTask.PREPARE_DATA_MESSAGE % datetime.datetime.now(): ''})

        cnt = 1
        for mo in report_helper.mo_list:
            # Не обрабатываем мо без дочерних ДОУ и
            # с признаком что они для фед. отчетности
            if not mo.get_descendants().filter(kind__id=UnitKind.DOU, gisdo__not_on_federal_report=False).exists():
                log.info(CollectUnitsDataTask.EMPTY_MO_MESSAGE % mo.display())
                continue
            has_error = False
            err_msg = None
            for dou_id in report_helper.dou_list[mo.id]:
                try:
                    dou = Unit.objects.get(id=dou_id)
                except Unit.DoesNotExist:
                    log.info(CollectUnitsDataTask.NOT_FOUND_MESSAGE % dou_id)
                    continue

                if dou.kind.id in [UnitKind.MO, UnitKind.REGION] and not dou.unit_set.count():
                    log.info(CollectUnitsDataTask.EMPTY_MO_MESSAGE % dou.display())
                    continue

                self.set_progress(u'(%s%%) Подсчет показателей по %s.' % (_get_report_progress(cnt, report_helper.unit_count),
                                                                          dou.display()))
                try:
                    report_helper.collect_data(dou)
                except Exception, e:
                    has_error = True
                    err_msg = e.message
                    break

                cnt += 1

            self._ui_log(mo.display(), has_error, err_msg)

        report_helper.report_form.in_progress = False
        report_helper.report_form.save()

        return self._result_message(u'Формирование завершено', state=states.SUCCESS)


class CollectUnitDataPeriodicTask(CollectUnitsDataTask):
    """
    Подсчет и отправка показателей (для периодического выполнения с celery beat)
    """
    description = u'Подсчет и отправка показателей'
    route_key = 'default'

    SUCCESS_SEND_MO_DATA = u'Данные по %s успешно отправлены'
    FAILURE_SEND_MO_DATA = u'Произошла ошибка при передаче данных по %s'

    def apply_async(self, args, kwargs, **options):
        options['routing_key'] = self.route_key
        result = super(CollectUnitDataPeriodicTask, self).apply_async(args, kwargs, **options)
        cur_user = get_current_user()
        RunningTask.objects.create(user=cur_user.get_profile() if cur_user else None,
                                   task_id=result.task_id,
                                   description=self.description)
        return result

    def _get_soap_client(self, report_helper):

        try:
            client = report_helper.create_client()
        except SendReportException, error:
            log.error(traceback.format_exc())
            raise SOAPClientCreateException(u'Ошибка при создании клиента. Сообщение: %s' % error.message)

        return client

    def run(self, **kwargs):
        """Выполнение задачи"""
        self.set_progress(u'Подсчет показателей...', task_state=u'RUNNING')

        user = get_current_user()

        highest_units = Unit.objects.filter(parent__isnull=True)
        if highest_units.count() == 1:
            unit = highest_units[0]
        else:
            raise ApplicationLogicException(u'В корне дерева учреждений должен быть только один элемент')

        report_form = ReportForm.objects.create(
            date=datetime.date.today(),
            user=user.get_profile() if user else None,
            unit=unit,
            in_progress=True)

        report_helper = ReportFormHelper(report_form, ScheduleSettings.get_settings())
        self.set_progress(values={CollectUnitDataPeriodicTask.PREPARE_DATA_MESSAGE % datetime.datetime.now(): ''})

        report_data = {
            'schema_version': '1.0',
            'report_date': report_helper.report_form.date.date().isoformat()
        }

        client = self._get_soap_client(report_helper)

        cnt = 1
        for mo in report_helper.mo_list:
            if not mo.unit_set.filter(gisdo__not_on_federal_report=False).count():
                log.info(CollectUnitsDataTask.EMPTY_MO_MESSAGE % mo.display())
                continue

            report_data['reports'] = []
            has_error = False
            err_msg = None
            for dou_id in report_helper.dou_list[mo.id]:
                try:
                    dou = Unit.objects.get(id=dou_id)
                except Unit.DoesNotExist:
                    log.info(CollectUnitsDataTask.NOT_FOUND_MESSAGE % dou_id)
                    continue

                if dou.kind.id in [UnitKind.MO] and not dou.unit_set.count():
                    log.info(CollectUnitsDataTask.EMPTY_MO_MESSAGE % dou.display())
                    continue
                progress = _get_report_progress(cnt, report_helper.unit_count)
                self.set_progress(u'(%s%%) Подсчет показателей по %s.' % (progress, dou.display()))

                try:
                    report_helper.collect_data(dou)
                except Exception, e:
                    has_error = True
                    err_msg = e.message
                    break

                self.set_progress(u'(%s%%) Генерация данных по %s.' % (progress, dou.display()))

                try:
                    sub_report = report_helper.create_sub_report(dou)
                except (ApplicationLogicException, KeyError), e:
                    has_error = True
                    err_msg = e.message
                    break

                report_data['reports'].append(sub_report)
                cnt += 1

            # По региону данные собираем, но не отправляем
            if mo.kind.id == UnitKind.REGION:
                continue

            report_data['send_time'] = datetime.datetime.now().isoformat()
            encoded_xml = report_helper.render_xml(report_data)

            self._ui_log(mo.display(), has_error, err_msg)
            # Если при сборе по МО произошли ошибки, то и отправлять ничего не надо.
            if has_error:
                continue

            try:
                report_helper.send(client, encoded_xml)
            except SendReportException, error:
                has_error = True
                err_msg = error.message
                report_helper.report_form.sent_time = None

            self._ui_log(mo.display(), has_error, err_msg,
                         success_message=CollectUnitDataPeriodicTask.SUCCESS_SEND_MO_DATA,
                         failure_message=CollectUnitDataPeriodicTask.FAILURE_SEND_MO_DATA)

        report_helper.report_form.in_progress = False
        report_helper.report_form.sent_time = datetime.datetime.now()
        report_helper.report_form.save()

        return self._result_message(u'Формирование завершено', state=states.SUCCESS)

    def on_failure(self, exc, task_id, args, kwargs, einfo):

        try:
            msg = unicode(exc.message)
        except UnicodeDecodeError:
            msg = exc.message.decode('UTF-8')

        self.set_progress(msg, task_state=states.FAILURE)